/*
    This file is part of Ansel
    Copyright (C) 2025 - Aurélien PIERRE

    Ansel is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Ansel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with darktable.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "develop/pixelpipe_cache.h"
#include "common/darktable.h"
#include "common/debug.h"
#include "develop/format.h"
#include "develop/pixelpipe_hb.h"
#include <glib.h>
#include <stdlib.h>


typedef struct dt_pixel_cache_entry_t
{
  uint64_t hash;
  void *data;
  size_t size;
  dt_iop_buffer_dsc_t dsc;
  int age;
  char *name; // name of the cache entry, for debugging
  int id;
} dt_pixel_cache_entry_t;


size_t dt_pixel_cache_get_size(dt_pixel_cache_entry_t *cache_entry)
{
  return cache_entry->size / (1024 * 1024);
}

void dt_pixel_cache_message(dt_pixel_cache_entry_t *cache_entry, const char *message)
{
  if(!((darktable.unmuted & DT_DEBUG_PIPE) && (darktable.unmuted & DT_DEBUG_VERBOSE))) return;
  dt_print(DT_DEBUG_PIPE, "[pixelpipe] cache entry %lu: %s (%lu MiB) %s\n", cache_entry->hash, cache_entry->name,
           dt_pixel_cache_get_size(cache_entry), message);
}

// remove the cache entry with the given hash and update the cache memory usage
void dt_dev_pixel_pipe_cache_remove(dt_dev_pixelpipe_cache_t *cache, const uint64_t hash)
{
  dt_pixel_cache_entry_t *cache_entry = (dt_pixel_cache_entry_t *)g_hash_table_lookup(cache->entries, GINT_TO_POINTER(hash));
  if(cache_entry)
  {
    cache->current_memory -= cache_entry->size;
    g_hash_table_remove(cache->entries, GINT_TO_POINTER(hash));
  }
  else
  {
    dt_print(DT_DEBUG_PIPE, "[pixelpipe] cache entry %lu not found, will not be removed\n", hash);
  }
}

typedef struct _cache_lru_t
{
  int max_age;
  uint64_t hash;
} _cache_lru_t;


// find the cache entry hash with the oldest use
void _cache_get_oldest(gpointer key, gpointer value, gpointer user_data)
{
  dt_pixel_cache_entry_t *cache_entry = (dt_pixel_cache_entry_t *)value;
  _cache_lru_t *lru = (_cache_lru_t *)user_data;
  if(cache_entry->age > lru->max_age)
  {
    lru->max_age = cache_entry->age;
    lru->hash = cache_entry->hash;
  }
}


// remove the least used cache entry
void dt_dev_pixel_pipe_cache_remove_lru(dt_dev_pixelpipe_cache_t *cache)
{
  _cache_lru_t *lru = (_cache_lru_t *)malloc(sizeof(_cache_lru_t));
  lru->max_age = -1;
  lru->hash = 0;
  g_hash_table_foreach(cache->entries, _cache_get_oldest, lru);

  // age = 0 means this is the most recently used entry, typically our output
  // age = 1 means this is the second most recently used entry, typically our input
  // we can't remove those without risk of segfaulting
  if(lru->max_age > 1)
    dt_dev_pixel_pipe_cache_remove(cache, lru->hash);

  free(lru);
}


static dt_pixel_cache_entry_t *dt_pixel_cache_new_entry(const uint64_t hash, const size_t size,
                                                        const dt_iop_buffer_dsc_t dsc, const char *name, const int id,
                                                        dt_dev_pixelpipe_cache_t *cache)
{
  // Free up space if needed to match the max memory limit
  while(cache->current_memory + size > cache->max_memory && g_hash_table_size(cache->entries) > 0)
    dt_dev_pixel_pipe_cache_remove_lru(cache);

  dt_pixel_cache_entry_t *cache_entry = (dt_pixel_cache_entry_t *)malloc(sizeof(dt_pixel_cache_entry_t));
  if(!cache_entry) return NULL;

  // allocate the data buffer
  cache_entry->data = dt_alloc_align(size);

  // if allocation failed, remove the least recently used cache entry, then try again
  while(cache_entry->data == NULL && g_hash_table_size(cache->entries) > 0)
  {
    dt_dev_pixel_pipe_cache_remove_lru(cache);
    cache_entry->data = dt_alloc_align(size);
  }

  if(!cache_entry->data)
  {
    free(cache_entry);
    return NULL;
  }

  cache_entry->size = size;
  cache_entry->age = 0;
  cache_entry->dsc = dsc;
  cache_entry->hash = hash;
  cache_entry->id = id;
  cache_entry->name = g_strdup(name);

  g_hash_table_insert(cache->entries, GINT_TO_POINTER(hash), cache_entry);
  cache->current_memory += size;

  return cache_entry;
}


static void _free_cache_entry(dt_pixel_cache_entry_t *cache_entry)
{
  if(!cache_entry) return;
  dt_pixel_cache_message(cache_entry, "freed");
  dt_free_align(cache_entry->data);
  cache_entry->data = NULL;
  g_free(cache_entry->name);
  free(cache_entry);
}


dt_dev_pixelpipe_cache_t * dt_dev_pixelpipe_cache_init(size_t max_memory)
{
  dt_dev_pixelpipe_cache_t *cache = (dt_dev_pixelpipe_cache_t *)malloc(sizeof(dt_dev_pixelpipe_cache_t));
  cache->entries = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, (GDestroyNotify)_free_cache_entry);
  cache->max_memory = max_memory;
  cache->current_memory = 0;
  cache->queries = cache->hits = 0;
  return cache;
}


void dt_dev_pixelpipe_cache_cleanup(dt_dev_pixelpipe_cache_t *cache)
{
  if(!cache) return;
  g_hash_table_destroy(cache->entries);
  cache->entries = NULL;
}


int dt_dev_pixelpipe_cache_available(dt_dev_pixelpipe_cache_t *cache, const uint64_t hash)
{
  return g_hash_table_lookup(cache->entries, GINT_TO_POINTER(hash)) != NULL;
}


void _age_cache_entry(gpointer key, gpointer value, gpointer user_data)
{
  dt_pixel_cache_entry_t *cache_entry = (dt_pixel_cache_entry_t *)value;
  cache_entry->age++;
}


int dt_dev_pixelpipe_cache_get(dt_dev_pixelpipe_cache_t *cache, const uint64_t hash,
                               const size_t size, const char *name, const int id,
                               void **data, dt_iop_buffer_dsc_t **dsc)
{
  cache->queries++;

  // Age all cache entries
  g_hash_table_foreach(cache->entries, _age_cache_entry, NULL);

  // Find the cache entry for this hash, if any
  dt_pixel_cache_entry_t *cache_entry = (dt_pixel_cache_entry_t *)g_hash_table_lookup(cache->entries, GINT_TO_POINTER(hash));
  gboolean cache_entry_found = (cache_entry != NULL);

  if(cache_entry)
    cache->hits++;
  else
    cache_entry = dt_pixel_cache_new_entry(hash, size, **dsc, name, id, cache);

  if(cache_entry)
  {
    cache_entry->age = 0; // this is the MRU entry
    *data = cache_entry->data;
    *dsc = &cache_entry->dsc;
    dt_pixel_cache_message(cache_entry, (cache_entry_found) ? "found" : "created");
    return !cache_entry_found;
  }
  else
  {
    *data = NULL;
    *dsc = NULL;
    dt_print(DT_DEBUG_PIPE, "couldn't allocate new cache entry %lu\n", hash);
    return 1;
  }
}

gboolean _for_each_remove(gpointer key, gpointer value, gpointer user_data)
{
  dt_pixel_cache_entry_t *cache_entry = (dt_pixel_cache_entry_t *)value;
  const int id = GPOINTER_TO_INT(user_data);
  return cache_entry->id == id || id == -1;
}

void dt_dev_pixelpipe_cache_flush(dt_dev_pixelpipe_cache_t *cache, const int id)
{
  g_hash_table_foreach_remove(cache->entries, _for_each_remove, GINT_TO_POINTER(id));
  cache->current_memory = 0;
}

typedef struct _cache_invalidate_t
{
  void *data;
  size_t size;
} _cache_invalidate_t;


gboolean _cache_invalidate(gpointer key, gpointer value, gpointer user_data)
{
  dt_pixel_cache_entry_t *cache_entry = (dt_pixel_cache_entry_t *)value;
  _cache_invalidate_t *cache_invalidate = (_cache_invalidate_t *)user_data;

  if(cache_entry->data == cache_invalidate->data)
  {
    cache_invalidate->size = cache_entry->size;
    return TRUE; // remove this entry
  }

  return FALSE; // keep this entry
}


void dt_dev_pixelpipe_cache_invalidate(dt_dev_pixelpipe_cache_t *cache, void *data)
{
  _cache_invalidate_t *cache_invalidate = (_cache_invalidate_t *)malloc(sizeof(_cache_invalidate_t));
  cache_invalidate->data = data;
  cache_invalidate->size = 0;

  // Remove the cache entry with the given data
  g_hash_table_foreach_remove(cache->entries, _cache_invalidate, cache_invalidate);

  // Update the current memory usage
  cache->current_memory -= cache_invalidate->size;

  free(cache_invalidate);
}


void dt_dev_pixelpipe_cache_print(dt_dev_pixelpipe_cache_t *cache)
{
  if(!(darktable.unmuted & DT_DEBUG_PIPE)) return;

  dt_print(DT_DEBUG_PIPE, "[pixelpipe] cache hit rate so far: %.3f%% - size: %lu MiB\n", 100. * (cache->hits) / (float)cache->queries, cache->current_memory / (1024 * 1024));
}

// clang-format off
// modelines: These editor modelines have been set for all relevant files by tools/update_modelines.py
// vim: shiftwidth=2 expandtab tabstop=2 cindent
// kate: tab-indents: off; indent-width 2; replace-tabs on; indent-mode cstyle; remove-trailing-spaces modified;
// clang-format on
